/***************************************************************************
                      serialtransport.c  -  description
                             -------------------
    begin                : Thu Dec 12 2002
    copyright            : (C) 2002 by James Stembridge
    email                : jstembridge@gmx.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <stdio.h>
#include <sys/poll.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <openobex/obex.h>

#include "serialtransport.h"


static void
serial_set_can(struct termios *options)
{
	options->c_lflag &= ~(ISIG | ECHO | ECHOE);
	options->c_lflag |= (ICANON | PENDIN);
	options->c_iflag &= ~(IXON | ICRNL | IEXTEN);
}


static void
serial_set_raw(struct termios *options)
{
	options->c_lflag &= ~(ICANON | ISIG | ECHO | ECHOE);
	options->c_lflag |= PENDIN;
}


static int
serial_read_to(int fd, unsigned char *buf, int count, int timeout)
{
	int ret;
	struct pollfd pfd;

	pfd.fd = fd;
	pfd.events = POLLIN;

	if((ret = poll(&pfd, 1, 1000*timeout)) < 0) {
		perror("serial_read_to: Error waiting for data from serial port");
		return -1;
	}
	if(ret == 0) {
		fprintf(stderr,
			"serial_read_to: Timeout while reading from serial port\n");
		return -1;
	}

	if(pfd.revents | POLLIN) {
		if((ret = read(fd, buf, count)) < 0) {
			perror("serial_read_to: Error reading from serial port");
			return -1;
		}
		return ret;
	}

	fprintf(stderr, "serial_read_to: Something bad happened\n");

	return -1;
}


static int
serial_at_cmd(int fd, const char *cmd, const char *rply, int timeout)
{
	unsigned char buffer[16];
	int nbytes;

	/* Find length of command */
    nbytes = strlen(cmd);

    /* Write command to serial port */
    if(write(fd, cmd, nbytes) < nbytes) {
        perror("serial_at_cmd: Error writing to serial port\n");
        return -1;
    }

    /* Read echo of command from phone */
    if((nbytes = serial_read_to(fd, buffer, sizeof(buffer), timeout)) < 0)
        return -1;

    /* Null terminate echo removing trailing \r\n */
    if(nbytes > 1)
        buffer[nbytes - 2] = '\0';

    /* Compare echo with sent command */
    if(strcmp(buffer, cmd) != 0) {
        fprintf(stderr, "serial_at_cmd: AT command echo error\n");
        return -1;
    }

	/* Read response to command from phone */
	if((nbytes = serial_read_to(fd, buffer, sizeof(buffer), timeout)) < 0)
		return -1;

	/* Null terminate response removing trailing \n */
	if(nbytes > 0)
		buffer[nbytes - 1] = '\0';

	/* Compare response with expected reply */
	if(strcmp(buffer, rply) != 0) {
		fprintf(stderr, "serial_at_cmd: AT commond reply error\n");
		return -1;
	}

	return 1;
}


int
serial_unset_hupcl(const char *device)
{
    int fd;
    struct termios options;

    /* Open serial port */
    if((fd = open(device, O_RDWR | O_NOCTTY | O_NDELAY, 0)) < 0) {
        perror("serial_unset_hupcl: Error opening serial port");
        return -1;
    }

    /* Get the current serial port options */
    tcgetattr(fd, &options);

    /* Unset HUPCL */
    options.c_cflag &= ~(HUPCL);

    /* Set options */
    tcsetattr(fd, TCSANOW, &options);

    /* Close serial port */
    if(close(fd) < 0) {
        perror("serial_unset_hupcl: Error closing serial port");
        return -1;
    }

    return 1;
}

    
int
serial_connect(obex_t *handle, void *userdata)
{
	struct termios options;
	struct serial_userdata *ud;
	int baud;

	(void) handle;

	ud = userdata;

    /* Attempt to open serial port */
	if((ud->fd = open(ud->device, O_RDWR | O_NOCTTY | O_NDELAY, 0)) < 0) {
		perror("serial_connect: Error opening serial port");
		return -1;
	}

	/* Get the current serial port options */
	tcgetattr(ud->fd, &options);

	/* Set baud rate */
	switch(ud->baud) {
		case 9600:
			baud = B9600;
			break;
		case 19200:
			baud = B19200;
			break;
		case 38400:
			baud = B38400;
			break;
		case 57600:
			baud = B57600;
			break;
		default:
			baud = B9600;
	}
	cfsetispeed(&options, baud);
	cfsetospeed(&options, baud);

	/* Set format to 8 data bits, 1 stop bit */
	options.c_cflag &= ~(CSTOPB | PARENB | CSIZE);
	options.c_cflag |= CS8;
	options.c_iflag |= IGNPAR;

	/* Enable local line, receiver, hardware flow control, raw output */
	options.c_cflag |= ( CLOCAL | CREAD | CRTSCTS);
	options.c_oflag &= ~(OPOST | OCRNL | ONLCR);

	/* Set for canonical input */
	serial_set_can(&options);

	/* Set options */
	tcflush(ud->fd, TCIOFLUSH);
	tcsetattr(ud->fd, TCSANOW, &options);

	/* Reset phone and put into OBEX mode */
	if((serial_at_cmd(ud->fd, "ATZ\r", "OK\r", 1) < 0) ||
       (serial_at_cmd(ud->fd, "AT*EOBEX\r", "CONNECT\r", 1) < 0)) {
    
        /* Connect has failed so we need to close serial port */
        if(close(ud->fd) < 0)
            perror("serial_connect: Error closing serial port");

        ud->fd = -1;
        return -1;
	}

	/* Set for raw input */
	serial_set_raw(&options);
	tcsetattr(ud->fd, TCSANOW, &options);

	return 1;
}


int
serial_disconnect(obex_t *handle, void *userdata)
{
	struct serial_userdata *ud;
	struct termios options;

	(void) handle;

	ud = userdata;

	/* Check we are actually connected */
	if(ud->fd == -1)
		return 1;

	/* Set for canonical input */
	tcgetattr(ud->fd, &options);
	serial_set_can(&options);
	tcsetattr(ud->fd, TCSANOW, &options);

	/* wait for OK after disconnect */
	if(serial_at_cmd(ud->fd, "", "OK\r", 100) < 0)
		return -1;

	/* Close serial port */
	if(close(ud->fd) < 0) {
		perror("serial_disconnect: Error closing serial port");
		return -1;
	}

	/* Reset fd */
	ud->fd = -1;

	return 1;
}


int
serial_write(obex_t *handle, void *userdata, uint8_t *buffer, int length)
{
	struct serial_userdata *ud;
	int nbytes;

	(void) handle;

    ud = userdata;

	if((nbytes = write(ud->fd, buffer, length)) < 0) {
		perror("serial_write: Error writing to serial port");
		return -1;
	}

	return nbytes;
}


int
serial_handleinput(obex_t *handle, void *userdata, int timeout)
{
	struct serial_userdata *ud;
	unsigned char buffer[32]; /* Arbitary size - find something better? */
	int nbytes;

	ud = userdata;

	/* Read data from serial port with timeout */
	if((nbytes = serial_read_to(ud->fd, buffer, sizeof(buffer), timeout)) < 0)
		return -1;
	
	if(nbytes == 0)
		return 0;

	/* Feed data to obex */
	OBEX_CustomDataFeed(handle, buffer, nbytes);

	return nbytes;
}


